# Copyright (C) 2019 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
from adi.attribute import attribute
from adi.context_manager import context_manager


class ad5686(context_manager, attribute):
    """ AD5686 DAC """

    _complex_data = False
    channel = []  # type: ignore
    _device_name = ""

    def __init__(self, uri="", device_index=0):
        context_manager.__init__(self, uri, self._device_name)
        # Dictionary with all compatible parts. The key of each entry is the device's id and it's value
        # is the number of bits the device supports.
        compatible_parts = [
            "ad5310r",
            "ad5311r",
            "ad5671r",
            "ad5672r",
            "ad5673r",
            "ad5674r",
            "ad5675r",
            "ad5676",
            "ad5676r",
            "ad5677r",
            "ad5679r",
            "ad5681r",
            "ad5682r",
            "ad5683",
            "ad5683r",
            "ad5684",
            "ad5684r",
            "ad5685r",
            "ad5686",
            "ad5686r",
            "ad5691r",
            "ad5692r",
            "ad5693",
            "ad5693r",
            "ad5694",
            "ad5694r",
            "ad5695r",
            "ad5696",
            "ad5696r",
        ]

        self._ctrl = None
        index = 0
        # We are selecting the device_index-th device from the 5686 family as working device.
        for device in self._ctx.devices:
            if device.name in compatible_parts:
                if index == device_index:
                    self._ctrl = device
                    break
                else:
                    index += 1

        for ch in self._ctrl.channels:
            name = ch.id
            self.channel.append(self._channel(self._ctrl, name))

        # sort device channels after the index of their index
        self.channel.sort(key=lambda x: int(x.name[7:]))

    class _channel(attribute):
        """AD5686 channel"""

        def __init__(self, ctrl, channel_name):
            self.name = channel_name
            self._ctrl = ctrl

        @property
        def raw(self):
            """AD5686 channel raw value"""
            return self._get_iio_attr(self.name, "raw", True, self._ctrl)

        @raw.setter
        def raw(self, value):
            self._set_iio_attr(self.name, "raw", True, str(int(value)))

        @property
        def powerdown(self):
            """AD5686 channel powerdown value"""
            return self._get_iio_attr(self.name, "powerdown", True)

        @powerdown.setter
        def powerdown(self, val):
            """AD5686 channel powerdown value"""
            self._set_iio_attr(self.name, "powerdown", True, val)

        @property
        def powerdown_mode(self):
            """AD5686 channel powerdown mode value"""
            return self._get_iio_attr_str(self.name, "powerdown_mode", True)

        @powerdown_mode.setter
        def powerdown_mode(self, val):
            """AD5686 channel powerdown value"""
            self._set_iio_attr_str(self.name, "powerdown_mode", True, val)

        @property
        def powerdown_mode_available(self):
            """Provides all available powerdown mode settings for the AD5686"""
            return self._get_iio_attr_str(self.name, "powerdown_mode_available", True)

        @property
        def scale(self):
            """AD5686 channel scale(gain)"""
            return self._get_iio_attr(self.name, "scale", True)

        def to_raw(self, val):
            """Converts raw value to SI"""
            return int(1000.0 * val / self.scale)

        @property
        def volts(self):
            """AD5686 channel value in volts"""
            return self.raw * self.scale

        @volts.setter
        def volts(self, val):
            """AD5686 channel value in volts"""
            self.raw = self.to_raw(val)



# Copyright (C) 2020 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from ctypes import c_char_p

import iio

from adi.context_manager import context_manager
from adi.obs import obs, remap, tx_two
from adi.rx_tx import rx_tx


def rx1(self):
    """rx1: Receive data on channel 0 (Same as rx() method) """
    return self.rx()


def rx2(self):
    """rx2: Receive data on channel 1 """
    return self._rx2.rx()


def tx1(self, data):
    """tx1: Transmit data on channel 0 (Same as tx() method) """
    self.tx(data)


def tx2(self, data):
    """tx2: Transmit data on channel 1 """
    self._tx2.tx(data)


class adrv9002(rx_tx, context_manager):
    """ ADRV9002 Transceiver """

    _complex_data = True
    _rx_channel_names = ["voltage0_i", "voltage0_q"]
    _rx2_channel_names = ["voltage0_i", "voltage0_q"]
    _tx_channel_names = ["voltage0", "voltage1"]
    _tx2_channel_names = ["voltage0", "voltage1"]
    _device_name = ""

    def __init__(self, uri=""):

        context_manager.__init__(self, uri, self._device_name)
        # Determine if we have a split or combined DMA
        devs = self._ctx.devices
        rxdevs = list(filter(lambda dev: "rx" in str(dev.name), devs))
        txdevs = list(filter(lambda dev: "tx" in str(dev.name), devs))

        if len(rxdevs) > 1:
            self._rx_dma_mode = "split"
            self._rxadc = self._ctx.find_device("axi-adrv9002-rx-lpc")
            self._rxadc2 = self._ctx.find_device("axi-adrv9002-rx2-lpc")
            self._rx2 = obs(self._ctx, self._rxadc2, self._rx2_channel_names)
            setattr(adrv9002, "rx1", rx1)
            setattr(adrv9002, "rx2", rx2)
            remap(self._rx2, "rx_", "rx2_", type(self))

        else:
            self._rx_dma_mode = "combined"
            self._rx_channel_names = [
                "voltage0_i",
                "voltage0_q",
                "voltage1_i",
                "voltage1_q",
            ]
            self._rxadc = self._ctx.find_device("axi-adrv9002-rx-lpc")

        if len(txdevs) > 1:
            self._tx_dma_mode = "split"
            self._txdac = self._ctx.find_device("axi-adrv9002-tx-lpc")
            self._txdac2 = self._ctx.find_device("axi-adrv9002-tx2-lpc")
            self._tx2 = tx_two(self._ctx, self._txdac2, self._tx2_channel_names)
            setattr(adrv9002, "tx1", tx1)
            setattr(adrv9002, "tx2", tx2)
            remap(self._tx2, "tx_", "tx2_", type(self))
            remap(self._tx2, "dds_", "dds2_", type(self))

        else:
            self._tx_dma_mode = "combined"
            self._tx_channel_names = ["voltage0", "voltage1", "voltage2", "voltage3"]
            self._txdac = self._ctx.find_device("axi-adrv9002-tx-lpc")

        self._ctrl = self._ctx.find_device("adrv9002-phy")

        self._ctx.set_timeout(30000)  # Needed for loading profiles

        rx_tx.__init__(self)

    def write_stream_profile(self, stream, profile):
        """Load a new profile and stream on the device"""
        with open(stream, "rb") as file:
            data = file.read()
            data = c_char_p(data)
            attr_encode = "stream_config".encode("ascii")
            iio._d_write_attr(self._ctrl._device, attr_encode, data)
        with open(profile, "r") as file:
            data = file.read()
        self._set_iio_dev_attr_str("profile_config", data)

    def write_profile(self, value):
        """Load a new profile on the device
            Stream related to profile should be loaded first.
            Please see driver documentation about profile generation.
        """
        with open(value, "r") as file:
            data = file.read()
        self._set_iio_dev_attr_str("profile_config", data)

    def write_stream(self, value):
        """Load a new stream on the device
            Stream becomes active once accompanying profile is loaded
            Please see driver documentation about stream generation.
        """
        with open(value, "rb") as file:
            data = file.read()
            data = c_char_p(data)
            attr_encode = "stream_config".encode("ascii")
            iio._d_write_attr(self._ctrl._device, attr_encode, data)

    # we cannot really get the profile. The driver will just throw EPERM
    profile = property(None, write_profile)
    stream = property(None, write_stream)

    @property
    def rx_dma_mode(self):
        """rx_dma_mode: DMA configuration for RX path. Options are:
            combined: RX1 and RX2 share the same rx method
            split: RX1 and RX2 have separate rx methods rx1 and rx2. Typically
            used when they are at different rates. In this case the standard rx
            method has the same effect as the rx1 method.
        """
        return self._rx_dma_mode

    @property
    def tx_dma_mode(self):
        """tx_dma_mode: DMA configuration for TX path. Options are:
                combined: TX1 and TX2 share the same tx method
                split: TX1 and TX2 have separate tx methods tx1 and tx2.
                Typically used when they are at different rates. In this case
                the standard tx method has the same effect as the tx1 method.
        """
        return self._tx_dma_mode

    @property
    def rx_ensm_mode_chan0(self):
        """rx_ensm_mode_chan0: RX Enable State Machine State Channel 0. Options are:
        calibrated, primed, rf_enabled"""
        return self._get_iio_attr_str("voltage0", "ensm_mode", False)

    @rx_ensm_mode_chan0.setter
    def rx_ensm_mode_chan0(self, value):
        self._set_iio_attr("voltage0", "ensm_mode", False, value)

    @property
    def rx_ensm_mode_chan1(self):
        """rx_ensm_mode_chan1: RX Enable State Machine State Channel 1. Options are:
        calibrated, primed, rf_enabled"""
        return self._get_iio_attr_str("voltage1", "ensm_mode", False)

    @rx_ensm_mode_chan1.setter
    def rx_ensm_mode_chan1(self, value):
        self._set_iio_attr("voltage1", "ensm_mode", False, value)

    @property
    def tx_ensm_mode_chan0(self):
        """tx_ensm_mode_chan0: TX Enable State Machine State Channel 0. Options are:
        calibrated, primed, rf_enabled"""
        return self._get_iio_attr_str("voltage0", "ensm_mode", True)

    @tx_ensm_mode_chan0.setter
    def tx_ensm_mode_chan0(self, value):
        self._set_iio_attr("voltage0", "ensm_mode", True, value)

    @property
    def tx_ensm_mode_chan1(self):
        """tx_ensm_mode_chan1: TX Enable State Machine State Channel 1. Options are:
        calibrated, primed, rf_enabled"""
        return self._get_iio_attr_str("voltage1", "ensm_mode", True)

    @tx_ensm_mode_chan1.setter
    def tx_ensm_mode_chan1(self, value):
        self._set_iio_attr("voltage1", "ensm_mode", True, value)

    @property
    def gain_control_mode_chan0(self):
        """gain_control_mode_chan0: Mode of receive path AGC. Options are:
        manual_spi, manual_pin, automatic"""
        return self._get_iio_attr_str("voltage0", "gain_control_mode", False)

    @gain_control_mode_chan0.setter
    def gain_control_mode_chan0(self, value):
        self._set_iio_attr("voltage0", "gain_control_mode", False, value)

    @property
    def gain_control_mode_chan1(self):
        """gain_control_mode_chan1: Mode of receive path AGC. Options are:
        manual_spi, manual_pin, automatic"""
        return self._get_iio_attr_str("voltage1", "gain_control_mode", False)

    @gain_control_mode_chan1.setter
    def gain_control_mode_chan1(self, value):
        self._set_iio_attr("voltage1", "gain_control_mode", False, value)

    @property
    def rx_hardwaregain_chan0(self):
        """rx_hardwaregain: Gain applied to RX path channel 0. Only applicable when
        gain_control_mode is set to 'spi'"""
        return self._get_iio_attr("voltage0", "hardwaregain", False)

    @rx_hardwaregain_chan0.setter
    def rx_hardwaregain_chan0(self, value):
        if self.gain_control_mode_chan0 == "spi":
            self._set_iio_attr("voltage0", "hardwaregain", False, value)

    @property
    def rx_hardwaregain_chan1(self):
        """rx_hardwaregain: Gain applied to RX path channel 1. Only applicable when
        gain_control_mode is set to 'spi'"""
        return self._get_iio_attr("voltage1", "hardwaregain", False)

    @rx_hardwaregain_chan1.setter
    def rx_hardwaregain_chan1(self, value):
        if self.gain_control_mode_chan1 == "spi":
            self._set_iio_attr("voltage1", "hardwaregain", False, value)

    @property
    def tx_hardwaregain_chan0(self):
        """tx_hardwaregain: Attenuation applied to TX path channel 0"""
        return self._get_iio_attr("voltage0", "hardwaregain", True)

    @tx_hardwaregain_chan0.setter
    def tx_hardwaregain_chan0(self, value):
        if self.atten_control_mode_chan0 == "spi":
            self._set_iio_attr("voltage0", "hardwaregain", True, value)

    @property
    def tx_hardwaregain_chan1(self):
        """tx_hardwaregain: Attenuation applied to TX path channel 1"""
        return self._get_iio_attr("voltage1", "hardwaregain", True)

    @tx_hardwaregain_chan1.setter
    def tx_hardwaregain_chan1(self, value):
        if self.atten_control_mode_chan1 == "spi":
            self._set_iio_attr("voltage1", "hardwaregain", True, value)

    @property
    def interface_gain_chan0(self):
        """interface_gain_chan0: Fixed input gain stage for channel 0.
        Options are: 18dB 12dB 6dB 0dB -6dB -12dB -18dB -24dB -30dB -36dB """
        return self._get_iio_attr_str("voltage0", "interface_gain", False)

    @interface_gain_chan0.setter
    def interface_gain_chan0(self, value):
        if (
            self.digital_gain_control_mode_chan0 == "spi"
        ) and self.rx_ensm_mode_chan0 == "rf_enabled":
            self._set_iio_attr("voltage0", "interface_gain", False, value)

    @property
    def interface_gain_chan1(self):
        """interface_gain_chan0: Fixed input gain stage for channel 0.
        Options are: 18dB 12dB 6dB 0dB -6dB -12dB -18dB -24dB -30dB -36dB """
        return self._get_iio_attr_str("voltage1", "interface_gain", False)

    @interface_gain_chan1.setter
    def interface_gain_chan1(self, value):
        if (
            self.digital_gain_control_mode_chan1 == "spi"
        ) and self.rx_ensm_mode_chan1 == "rf_enabled":
            self._set_iio_attr("voltage1", "interface_gain", False, value)

    @property
    def agc_tracking_en_chan0(self):
        """Enable AGC on the fly tracking calibration for RX1"""
        return self._get_iio_attr("voltage0", "agc_tracking_en", False)

    @agc_tracking_en_chan0.setter
    def agc_tracking_en_chan0(self, value):
        self._set_iio_attr("voltage0", "agc_tracking_en", False, value)

    @property
    def bbdc_rejection_tracking_en_chan0(self):
        """"Enable Baseband DC rejection on the fly tracking calibration for RX1"""
        return self._get_iio_attr("voltage0", "bbdc_rejection_tracking_en", False)

    @bbdc_rejection_tracking_en_chan0.setter
    def bbdc_rejection_tracking_en_chan0(self, value):
        self._set_iio_attr("voltage0", "bbdc_rejection_tracking_en", False, value)

    @property
    def hd_tracking_en_chan0(self):
        """"Enable Harmonic Distortion on the fly tracking calibration for RX1"""
        return self._get_iio_attr("voltage0", "hd_tracking_en", False)

    @hd_tracking_en_chan0.setter
    def hd_tracking_en_chan0(self, value):
        self._set_iio_attr("voltage0", "hd_tracking_en", False, value)

    @property
    def quadrature_fic_tracking_en_chan0(self):
        """Enable Quadrature Error Correction Narrowband FIC on the fly tracking
        calibration for RX1"""
        return self._get_iio_attr("voltage0", "quadrature_fic_tracking_en", False)

    @quadrature_fic_tracking_en_chan0.setter
    def quadrature_fic_tracking_en_chan0(self, value):
        self._set_iio_attr("voltage0", "quadrature_fic_tracking_en", False, value)

    @property
    def quadrature_w_poly_tracking_en_chan0(self):
        """Enable Quadrature Error Correction Wideband Poly on the fly tracking
        calibration for RX1"""
        return self._get_iio_attr("voltage0", "quadrature_w_poly_tracking_en", False)

    @quadrature_w_poly_tracking_en_chan0.setter
    def quadrature_w_poly_tracking_en_chan0(self, value):
        self._set_iio_attr("voltage0", "quadrature_w_poly_tracking_en", False, value)

    @property
    def rfdc_tracking_en_chan0(self):
        """"Enable RF DC on the fly tracking calibration for RX1"""
        return self._get_iio_attr("voltage0", "rfdc_tracking_en", False)

    @rfdc_tracking_en_chan0.setter
    def rfdc_tracking_en_chan0(self, value):
        self._set_iio_attr("voltage0", "rfdc_tracking_en", False, value)

    @property
    def rssi_tracking_en_chan0(self):
        """"Enable RSSI on the fly tracking calibration for RX1"""
        return self._get_iio_attr("voltage0", "rssi_tracking_en", False)

    @rssi_tracking_en_chan0.setter
    def rssi_tracking_en_chan0(self, value):
        self._set_iio_attr("voltage0", "rssi_tracking_en", False, value)

    @property
    def agc_tracking_en_chan1(self):
        """Enable AGC on the fly tracking calibration for RX2"""
        return self._get_iio_attr("voltage1", "agc_tracking_en", False)

    @agc_tracking_en_chan1.setter
    def agc_tracking_en_chan1(self, value):
        self._set_iio_attr("voltage1", "agc_tracking_en", False, value)

    @property
    def bbdc_rejection_tracking_en_chan1(self):
        """"Enable Baseband DC rejection on the fly tracking calibration for RX2"""
        return self._get_iio_attr("voltage1", "bbdc_rejection_tracking_en", False)

    @bbdc_rejection_tracking_en_chan1.setter
    def bbdc_rejection_tracking_en_chan1(self, value):
        self._set_iio_attr("voltage1", "bbdc_rejection_tracking_en", False, value)

    @property
    def hd_tracking_en_chan1(self):
        """"Enable Harmonic Distortion on the fly tracking calibration for RX2"""
        return self._get_iio_attr("voltage1", "hd_tracking_en", False)

    @hd_tracking_en_chan1.setter
    def hd_tracking_en_chan1(self, value):
        self._set_iio_attr("voltage1", "hd_tracking_en", False, value)

    @property
    def quadrature_fic_tracking_en_chan1(self):
        """Enable Quadrature Error Correction Narrowband FIC on the fly tracking
        calibration for RX2"""
        return self._get_iio_attr("voltage1", "quadrature_fic_tracking_en", False)

    @quadrature_fic_tracking_en_chan1.setter
    def quadrature_fic_tracking_en_chan1(self, value):
        self._set_iio_attr("voltage1", "quadrature_fic_tracking_en", False, value)

    @property
    def quadrature_w_poly_tracking_en_chan1(self):
        """Enable Quadrature Error Correction Wideband Poly on the fly tracking
        calibration for RX2"""
        return self._get_iio_attr("voltage1", "quadrature_w_poly_tracking_en", False)

    @quadrature_w_poly_tracking_en_chan1.setter
    def quadrature_w_poly_tracking_en_chan1(self, value):
        self._set_iio_attr("voltage1", "quadrature_w_poly_tracking_en", False, value)

    @property
    def rfdc_tracking_en_chan1(self):
        """"Enable RF DC on the fly tracking calibration for RX2"""
        return self._get_iio_attr("voltage1", "rfdc_tracking_en", False)

    @rfdc_tracking_en_chan1.setter
    def rfdc_tracking_en_chan1(self, value):
        self._set_iio_attr("voltage1", "rfdc_tracking_en", False, value)

    @property
    def rssi_tracking_en_chan1(self):
        """"Enable RSSI on the fly tracking calibration for RX2"""
        return self._get_iio_attr("voltage1", "rssi_tracking_en", False)

    @rssi_tracking_en_chan1.setter
    def rssi_tracking_en_chan1(self, value):
        self._set_iio_attr("voltage1", "rssi_tracking_en", False, value)

    @property
    def close_loop_gain_tracking_en_chan0(self):
        """Enable Close Loop Gain tracking calibration for TX1"""
        return self._get_iio_attr("voltage0", "close_loop_gain_tracking_en", True)

    @close_loop_gain_tracking_en_chan0.setter
    def close_loop_gain_tracking_en_chan0(self, value):
        self._set_iio_attr("voltage0", "close_loop_gain_tracking_en", True, value)

    @property
    def lo_leakage_tracking_en_chan0(self):
        """Enable LO Leakage tracking calibration for TX1"""
        return self._get_iio_attr("voltage0", "lo_leakage_tracking_en", True)

    @lo_leakage_tracking_en_chan0.setter
    def lo_leakage_tracking_en_chan0(self, value):
        self._set_iio_attr("voltage0", "lo_leakage_tracking_en", True, value)

    @property
    def loopback_delay_tracking_en_chan0(self):
        """Enable Loopback delay tracking calibration for TX1"""
        return self._get_iio_attr("voltage0", "loopback_delay_tracking_en", True)

    @loopback_delay_tracking_en_chan0.setter
    def loopback_delay_tracking_en_chan0(self, value):
        self._set_iio_attr("voltage0", "loopback_delay_tracking_en", True, value)

    @property
    def pa_correction_tracking_en_chan0(self):
        """Enable PA Correction tracking calibration for TX1"""
        return self._get_iio_attr("voltage0", "pa_correction_tracking_en", True)

    @pa_correction_tracking_en_chan0.setter
    def pa_correction_tracking_en_chan0(self, value):
        self._set_iio_attr("voltage0", "pa_correction_tracking_en", True, value)

    @property
    def quadrature_tracking_en_chan0(self):
        """Enable Quadrature tracking calibration for TX1"""
        return self._get_iio_attr("voltage0", "quadrature_tracking_en", True)

    @quadrature_tracking_en_chan0.setter
    def quadrature_tracking_en_chan0(self, value):
        self._set_iio_attr("voltage0", "quadrature_tracking_en", True, value)

    @property
    def close_loop_gain_tracking_en_chan1(self):
        """Enable Close Loop Gain tracking calibration for TX2"""
        return self._get_iio_attr("voltage1", "close_loop_gain_tracking_en", True)

    @close_loop_gain_tracking_en_chan1.setter
    def close_loop_gain_tracking_en_chan1(self, value):
        self._set_iio_attr("voltage1", "close_loop_gain_tracking_en", True, value)

    @property
    def lo_leakage_tracking_en_chan1(self):
        """Enable LO Leakage tracking calibration for TX2"""
        return self._get_iio_attr("voltage1", "lo_leakage_tracking_en", True)

    @lo_leakage_tracking_en_chan1.setter
    def lo_leakage_tracking_en_chan1(self, value):
        self._set_iio_attr("voltage1", "lo_leakage_tracking_en", True, value)

    @property
    def loopback_delay_tracking_en_chan1(self):
        """Enable Loopback delay tracking calibration for TX2"""
        return self._get_iio_attr("voltage1", "loopback_delay_tracking_en", True)

    @loopback_delay_tracking_en_chan1.setter
    def loopback_delay_tracking_en_chan1(self, value):
        self._set_iio_attr("voltage1", "loopback_delay_tracking_en", True, value)

    @property
    def pa_correction_tracking_en_chan1(self):
        """Enable PA Correction tracking calibration for TX2"""
        return self._get_iio_attr("voltage1", "pa_correction_tracking_en", True)

    @pa_correction_tracking_en_chan1.setter
    def pa_correction_tracking_en_chan1(self, value):
        self._set_iio_attr("voltage1", "pa_correction_tracking_en", True, value)

    @property
    def quadrature_tracking_en_chan1(self):
        """Enable Quadrature tracking calibration for TX2"""
        return self._get_iio_attr("voltage1", "quadrature_tracking_en", True)

    @quadrature_tracking_en_chan1.setter
    def quadrature_tracking_en_chan1(self, value):
        self._set_iio_attr("voltage1", "quadrature_tracking_en", True, value)

    @property
    def digital_gain_control_mode_chan0(self):
        """Digital gain control mode for RX1. Option are: automatic spi."""
        return self._get_iio_attr_str("voltage0", "digital_gain_control_mode", False)

    @digital_gain_control_mode_chan0.setter
    def digital_gain_control_mode_chan0(self, value):
        self._set_iio_attr("voltage0", "digital_gain_control_mode", False, value)

    @property
    def digital_gain_control_mode_chan1(self):
        """ Digital gain control mode for RX2. Option are: automatic spi."""
        return self._get_iio_attr_str("voltage1", "digital_gain_control_mode", False)

    @digital_gain_control_mode_chan1.setter
    def digital_gain_control_mode_chan1(self, value):
        self._set_iio_attr("voltage1", "digital_gain_control_mode", False, value)

    @property
    def rx0_port_en(self):
        """Control Port RF Enable mode for RX1. Options are: pin and spi"""
        return self._get_iio_attr_str("voltage0", "port_en_mode", False)

    @rx0_port_en.setter
    def rx0_port_en(self, value):
        self._set_iio_attr("voltage0", "port_en_mode", False, value)

    @property
    def tx0_port_en(self):
        """Control Port RF Enable mode for TX1. Options are: pin and spi"""
        return self._get_iio_attr_str("voltage0", "port_en_mode", True)

    @tx0_port_en.setter
    def tx0_port_en(self, value):
        self._set_iio_attr("voltage0", "port_en_mode", True, value)

    @property
    def rx1_port_en(self):
        """Control Port RF Enable mode for RX2. Options are: pin and spi"""
        return self._get_iio_attr_str("voltage1", "port_en_mode", False)

    @rx1_port_en.setter
    def rx1_port_en(self, value):
        self._set_iio_attr("voltage1", "port_en_mode", False, value)

    @property
    def tx1_port_en(self):
        """Control Port RF Enable mode for TX2. Options are: pin and spi"""
        return self._get_iio_attr_str("voltage1", "port_en_mode", True)

    @tx1_port_en.setter
    def tx1_port_en(self, value):
        self._set_iio_attr("voltage1", "port_en_mode", True, value)

    @property
    def rx0_en(self):
        """Control RX1 Power state"""
        return self._get_iio_attr("voltage0", "en", False)

    @rx0_en.setter
    def rx0_en(self, value):
        self._set_iio_attr("voltage0", "en", False, value)

    @property
    def tx0_en(self):
        """"Control TX1 Power state"""
        return self._get_iio_attr("voltage0", "en", True)

    @tx0_en.setter
    def tx0_en(self, value):
        self._set_iio_attr("voltage0", "en", True, value)

    @property
    def rx1_en(self):
        """"Control RX2 Power state"""
        return self._get_iio_attr("voltage1", "en", False)

    @rx1_en.setter
    def rx1_en(self, value):
        self._set_iio_attr("voltage1", "en", False, value)

    @property
    def tx1_en(self):
        """"Control TX2 Power state"""
        return self._get_iio_attr("voltage1", "en", True)

    @tx1_en.setter
    def tx1_en(self, value):
        self._set_iio_attr("voltage1", "en", True, value)

    @property
    def rx0_nco_frequency(self):
        """NCO correction frequency for RX1"""
        return self._get_iio_attr("voltage0", "nco_frequency", False)

    @rx0_nco_frequency.setter
    def rx0_nco_frequency(self, value):
        self._set_iio_attr("voltage0", "nco_frequency", False, value)

    @property
    def tx0_nco_frequency(self):
        """NCO correction frequency for TX1"""
        return self._get_iio_attr("voltage0", "nco_frequency", True)

    @tx0_nco_frequency.setter
    def tx0_nco_frequency(self, value):
        self._set_iio_attr("voltage0", "nco_frequency", True, value)

    @property
    def rx1_nco_frequency(self):
        """NCO correction frequency for RX2"""
        return self._get_iio_attr("voltage1", "nco_frequency", False)

    @rx1_nco_frequency.setter
    def rx1_nco_frequency(self, value):
        self._set_iio_attr("voltage1", "nco_frequency", False, value)

    @property
    def tx1_nco_frequency(self):
        """NCO correction frequency for TX2"""
        return self._get_iio_attr("voltage1", "nco_frequency", True)

    @tx1_nco_frequency.setter
    def tx1_nco_frequency(self, value):
        self._set_iio_attr("voltage1", "nco_frequency", True, value)

    @property
    def atten_control_mode_chan0(self):
        """Control TX1 attenuation mode. Options are: bypass, spi, pin"""
        return self._get_iio_attr_str("voltage0", "atten_control_mode", True)

    @atten_control_mode_chan0.setter
    def atten_control_mode_chan0(self, value):
        self._set_iio_attr("voltage0", "atten_control_mode", True, value)

    @property
    def atten_control_mode_chan1(self):
        """Control TX2 attenuation mode. Options are: bypass, spi, pin"""
        return self._get_iio_attr_str("voltage1", "atten_control_mode", True)

    @atten_control_mode_chan1.setter
    def atten_control_mode_chan1(self, value):
        self._set_iio_attr("voltage1", "atten_control_mode", True, value)

    @property
    def rx0_rf_bandwidth(self):
        """rx_rf_bandwidth: Bandwidth of front-end analog filter of RX1 path"""
        return self._get_iio_attr("voltage0", "rf_bandwidth", False)

    @property
    def tx0_rf_bandwidth(self):
        """tx_rf_bandwidth: Bandwidth of front-end analog filter of TX1 path"""
        return self._get_iio_attr("voltage0", "rf_bandwidth", True)

    @property
    def rx0_sample_rate(self):
        """rx_sample_rate: Sample rate RX1 path in samples per second"""
        return self._get_iio_attr("voltage0", "sampling_frequency", False)

    @property
    def tx0_sample_rate(self):
        """tx_sample_rate: Sample rate TX1 path in samples per second"""
        return self._get_iio_attr("voltage0", "sampling_frequency", True)

    @property
    def rx1_rf_bandwidth(self):
        """rx_rf_bandwidth: Bandwidth of front-end analog filter of RX2 path"""
        return self._get_iio_attr("voltage1", "rf_bandwidth", False)

    @property
    def tx1_rf_bandwidth(self):
        """tx_rf_bandwidth: Bandwidth of front-end analog filter of TX2 path"""
        return self._get_iio_attr("voltage1", "rf_bandwidth", True)

    @property
    def rx1_sample_rate(self):
        """rx_sample_rate: Sample rate RX2 path in samples per second"""
        return self._get_iio_attr("voltage1", "sampling_frequency", False)

    @property
    def tx1_sample_rate(self):
        """tx_sample_rate: Sample rate TX2 path in samples per second"""
        return self._get_iio_attr("voltage1", "sampling_frequency", True)

    @property
    def rx0_lo(self):
        """rx0_lo: Carrier frequency of RX1 path"""
        return self._get_iio_attr("altvoltage0", "RX1_LO_frequency", True)

    @rx0_lo.setter
    def rx0_lo(self, value):
        self._set_iio_attr("altvoltage0", "RX1_LO_frequency", True, value)

    @property
    def rx1_lo(self):
        """rx1_lo: Carrier frequency of RX2 path"""
        return self._get_iio_attr("altvoltage1", "RX2_LO_frequency", True)

    @rx1_lo.setter
    def rx1_lo(self, value):
        self._set_iio_attr("altvoltage1", "RX2_LO_frequency", True, value)

    @property
    def tx0_lo(self):
        """tx1_lo: Carrier frequency of TX1 path"""
        return self._get_iio_attr("altvoltage2", "TX1_LO_frequency", True)

    @tx0_lo.setter
    def tx0_lo(self, value):
        self._set_iio_attr("altvoltage2", "TX1_LO_frequency", True, value)

    @property
    def tx1_lo(self):
        """tx1_lo: Carrier frequency of TX2 path"""
        return self._get_iio_attr("altvoltage3", "TX2_LO_frequency", True)

    @tx1_lo.setter
    def tx1_lo(self, value):
        self._set_iio_attr("altvoltage3", "TX2_LO_frequency", True, value)



import numbers
from collections import OrderedDict
from collections.abc import Iterable

import numpy as np
from adi.attribute import attribute
from adi.context_manager import context_manager
from adi.rx_tx import rx


class ltc2983(rx, context_manager):
    """ LTC2983 Multi-Sensor Temperature Measurement System """

    channel: OrderedDict = None
    _device_name = "ltc2983"
    _rx_unbuffered_data = True
    _rx_data_type = np.int32
    _rx_data_si_type = float

    def __init__(self, uri=""):
        context_manager.__init__(self, uri, self._device_name)
        self._ctrl = self._ctx.find_device("ltc2983")
        self._rxadc = self._ctx.find_device("ltc2983")

        # dynamically get channels
        _channels = []
        for ch in self._ctrl.channels:
            self._rx_channel_names.append(ch.id)
            _channels.append((ch.id, self._channel(self._ctrl, ch.id)))
        self.channel = OrderedDict(_channels)

        rx.__init__(self)

    class _channel(attribute):
        """ LTC2983 channel """

        def __init__(self, ctrl, channel_name):
            self._ctrl = ctrl
            self.name = channel_name

        @property
        def raw(self):
            """Channel raw value"""
            return self._get_iio_attr(self.name, "raw", False)

        @property
        def scale(self):
            """Channel scale factor"""
            return float(self._get_iio_attr_str(self.name, "scale", False))

        @property
        def value(self):
            """Value in real units"""
            return self.raw * self.scale

    def convert(self, channel_name, val):
        """Convert raw value(s) to real units"""
        if isinstance(channel_name, numbers.Integral):
            # self.channel is ordered
            channel_name = list(self.channel.keys())[channel_name]

        if isinstance(val, Iterable):
            # don't copy unless really needed
            try:
                val = np.asarray(val, np.int32)
            except TypeError:
                val = np.fromiter(val, np.int32)

        return val * self.channel[channel_name].scale



# Copyright (C) 2022 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import numpy as np
from adi.attribute import attribute
from adi.context_manager import context_manager
from adi.rx_tx import rx


class max31865(rx, context_manager, attribute):
    """MAX31865 RTD to Digital device"""

    _device_name = "max31865"
    _rx_data_type = np.int16
    _rx_unbuffered_data = True
    _rx_data_si_type = float

    def __init__(self, uri=""):
        context_manager.__init__(self, uri, self._device_name)
        self._ctrl = self._ctx.find_device("max31865")

        if self._ctrl is None:
            raise Exception("No device found")

        self.temp = self._channel_temp(self._ctrl, "temp")
        self.filter = self._channel_filt(self._ctrl, "filter")
        self._rx_channel_names = ["temp", "filter"]
        rx.__init__(self)

    @property
    def fault(self):
        """MAX31865 Over/Under Voltage Flag"""
        return self._get_iio_dev_attr("fault_ovuv")

    @property
    def samp_available(self):
        """MAX31865 Sampling frequency"""
        return self._get_iio_dev_attr("sampling_frequency_available")

    class _channel_temp(attribute):
        """MAX31865 temp channel"""

        def __init__(self, ctrl, channel_name):
            self.name = channel_name
            self._ctrl = ctrl

        @property
        def raw(self):
            """MAX31865 channel raw value"""
            return self._get_iio_attr(self.name, "raw", False)

        @property
        def scale(self):
            """MAX31865 channel scale value"""
            return self._get_iio_attr(self.name, "scale", False)

    class _channel_filt(attribute):
        """MAX31865 filter channel"""

        def __init__(self, ctrl, channel_name):
            self.name = channel_name
            self._ctrl = ctrl

        @property
        def notch(self):
            """MAX31865 Notch center frequency"""
            return self._get_iio_attr(self.name, "notch_center_frequency", False)

        @notch.setter
        def notch(self, value):
            """MAX31865 Notch center frequency (50 or 60)"""
            return self._set_iio_attr(self.name, "notch_center_frequency", False, value)



# Copyright (C) 2020 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import time

from adi.cn0540 import cn0540, reset_buffer


class cn0532(cn0540):
    """CN0532: Custom board with ADXL1002 Low Noise, High Frequency +/-50g MEMS Accelerometer"""

    def __init__(self, uri=""):

        cn0540.__init__(self, uri=uri)

    @reset_buffer
    def calibrate(self):
        """Tune LTC2606 to make AD7768-1 ADC codes zero mean"""
        adc_chan = self._rxadc
        dac_chan = self._ltc2606
        adc_scale = float(self._get_iio_attr("voltage0", "scale", False, adc_chan))
        dac_scale = float(self._get_iio_attr("voltage0", "scale", True, dac_chan))

        for _ in range(20):
            raw = self._get_iio_attr("voltage0", "raw", False, adc_chan)
            adc_voltage = raw * adc_scale

            raw = self._get_iio_attr("voltage0", "raw", True, dac_chan)
            dac_voltage = (raw * dac_scale - adc_voltage) / dac_scale

            e = int(dac_voltage * dac_scale)

            if int(dac_voltage) > 2 ** 16 - 1:
                print(
                    "Warning: DAC voltage at upper limit, "
                    + f"calibration may not converge (Error: {e-(2**16-1)} codes).\n"
                    + "Make sure sensor is connected."
                )
                dac_voltage = 2 ** 16 - 1
            elif int(dac_voltage) < 0:
                print(
                    "Warning: DAC voltage at lower limit, "
                    + f"calibration may not converge (Error: {e} codes).\n"
                    + "Make sure sensor is connected."
                )
                dac_voltage = 0

            self._set_iio_attr_float(
                "voltage0", "raw", True, int(dac_voltage), dac_chan
            )
            time.sleep(0.01)



# Copyright (C) 2019 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from adi.context_manager import context_manager
from adi.rx_tx import rx_tx_def

from .dsp import _dec_int_fpga_filter


class ad9364(rx_tx_def, context_manager):
    """AD9364 Transceiver"""

    _complex_data = True
    _rx_channel_names = ["voltage0", "voltage1"]
    _tx_channel_names = ["voltage0", "voltage1"]
    _control_device_name = "ad9361-phy"
    _rx_data_device_name = "cf-ad9361-lpc"
    _tx_data_device_name = "cf-ad9361-dds-core-lpc"
    _device_name = ""

    @property
    def filter(self):
        """Load FIR filter file. Provide path to filter file to attribute"""
        return self._get_iio_dev_attr("filter_fir_config")

    @filter.setter
    def filter(self, value):
        with open(value, "r") as file:
            data = file.read()
        self.sample_rate = 3000000
        self._set_iio_attr("out", "voltage_filter_fir_en", False, 0)
        self._set_iio_dev_attr_str("filter_fir_config", data)
        self._set_iio_attr("out", "voltage_filter_fir_en", False, 1)

    @property
    def loopback(self):
        """loopback: Set loopback mode. Options are:
        0 (Disable), 1 (Digital), 2 (RF)"""
        return self._get_iio_debug_attr("loopback")

    @loopback.setter
    def loopback(self, value):
        self._set_iio_debug_attr_str("loopback", value)

    @property
    def gain_control_mode_chan0(self):
        """gain_control_mode_chan0: Mode of receive path AGC. Options are:
        slow_attack, fast_attack, manual"""
        return self._get_iio_attr_str("voltage0", "gain_control_mode", False)

    @gain_control_mode_chan0.setter
    def gain_control_mode_chan0(self, value):
        self._set_iio_attr("voltage0", "gain_control_mode", False, value)

    @property
    def rx_hardwaregain_chan0(self):
        """rx_hardwaregain_chan0: Gain applied to RX path. Only applicable when
        gain_control_mode is set to 'manual'"""
        return self._get_iio_attr("voltage0", "hardwaregain", False)

    @rx_hardwaregain_chan0.setter
    def rx_hardwaregain_chan0(self, value):
        if self.gain_control_mode_chan0 == "manual":
            self._set_iio_attr_float("voltage0", "hardwaregain", False, value)

    @property
    def tx_hardwaregain_chan0(self):
        """tx_hardwaregain_chan0: Attenuation applied to TX path"""
        return self._get_iio_attr("voltage0", "hardwaregain", True)

    @tx_hardwaregain_chan0.setter
    def tx_hardwaregain_chan0(self, value):
        self._set_iio_attr_float("voltage0", "hardwaregain", True, value)

    @property
    def rx_rf_bandwidth(self):
        """rx_rf_bandwidth: Bandwidth of front-end analog filter of RX path"""
        return self._get_iio_attr("voltage0", "rf_bandwidth", False)

    @rx_rf_bandwidth.setter
    def rx_rf_bandwidth(self, value):
        self._set_iio_attr_int("voltage0", "rf_bandwidth", False, value)

    @property
    def tx_rf_bandwidth(self):
        """tx_rf_bandwidth: Bandwidth of front-end analog filter of TX path"""
        return self._get_iio_attr("voltage0", "rf_bandwidth", True)

    @tx_rf_bandwidth.setter
    def tx_rf_bandwidth(self, value):
        self._set_iio_attr_int("voltage0", "rf_bandwidth", True, value)

    @property
    def sample_rate(self):
        """sample_rate: Sample rate RX and TX paths in samples per second"""
        return self._get_iio_attr("voltage0", "sampling_frequency", False)

    @sample_rate.setter
    def sample_rate(self, rate):
        if rate < 521e3:
            raise ValueError(
                "Error: Does not currently support sample rates below 521e3"
            )

        # The following was converted from ad9361_set_bb_rate() in libad9361-iio
        # fmt: off
        if (rate <= 20000000):
            dec = 4
            fir = [
                -15, -27, -23, -6, 17, 33, 31, 9, -23, -47, -45, -13, 34, 69,
                67, 21, -49, -102, -99, -32, 69, 146, 143, 48, -96, -204, -200,
                -69, 129, 278, 275, 97, -170, -372, -371, -135, 222, 494, 497,
                187, -288, -654, -665, -258, 376, 875, 902, 363, -500, -1201,
                -1265, -530, 699, 1748, 1906, 845, -1089, -2922, -3424, -1697,
                2326, 7714, 12821, 15921, 15921, 12821, 7714, 2326, -1697,
                -3424, -2922, -1089, 845, 1906, 1748, 699, -530, -1265, -1201,
                -500, 363, 902, 875, 376, -258, -665, -654, -288, 187, 497,
                494, 222, -135, -371, -372, -170, 97, 275, 278, 129, -69, -200,
                -204, -96, 48, 143, 146, 69, -32, -99, -102, -49, 21, 67, 69,
                34, -13, -45, -47, -23, 9, 31, 33, 17, -6, -23, -27, -15
            ]
            taps = 128
        elif (rate <= 40000000):
            dec = 2
            fir = [
                -0, 0, 1, -0, -2, 0, 3, -0, -5, 0, 8, -0, -11, 0, 17, -0, -24,
                0, 33, -0, -45, 0, 61, -0, -80, 0, 104, -0, -134, 0, 169, -0,
                -213, 0, 264, -0, -327, 0, 401, -0, -489, 0, 595, -0, -724, 0,
                880, -0, -1075, 0, 1323, -0, -1652, 0, 2114, -0, -2819, 0,
                4056, -0, -6883, 0, 20837, 32767, 20837, 0, -6883, -0, 4056, 0,
                -2819, -0, 2114, 0, -1652, -0, 1323, 0, -1075, -0, 880, 0,
                -724, -0, 595, 0, -489, -0, 401, 0, -327, -0, 264, 0, -213, -0,
                169, 0, -134, -0, 104, 0, -80, -0, 61, 0, -45, -0, 33, 0, -24,
                -0, 17, 0, -11, -0, 8, 0, -5, -0, 3, 0, -2, -0, 1, 0, -0, 0
            ]
            taps = 128
        elif (rate <= 53333333):
            dec = 2
            fir = [
                -4, 0, 8, -0, -14, 0, 23, -0, -36, 0, 52, -0, -75, 0, 104, -0,
                -140, 0, 186, -0, -243, 0, 314, -0, -400, 0, 505, -0, -634, 0,
                793, -0, -993, 0, 1247, -0, -1585, 0, 2056, -0, -2773, 0, 4022,
                -0, -6862, 0, 20830, 32767, 20830, 0, -6862, -0, 4022, 0,
                -2773, -0, 2056, 0, -1585, -0, 1247, 0, -993, -0, 793, 0, -634,
                -0, 505, 0, -400, -0, 314, 0, -243, -0, 186, 0, -140, -0, 104,
                0, -75, -0, 52, 0, -36, -0, 23, 0, -14, -0, 8, 0, -4, 0
            ]
            taps = 96
        else:
            dec = 2
            fir = [
                -58, 0, 83, -0, -127, 0, 185, -0, -262, 0, 361, -0, -488, 0,
                648, -0, -853, 0, 1117, -0, -1466, 0, 1954, -0, -2689, 0, 3960,
                -0, -6825, 0, 20818, 32767, 20818, 0, -6825, -0, 3960, 0,
                -2689, -0, 1954, 0, -1466, -0, 1117, 0, -853, -0, 648, 0, -488,
                -0, 361, 0, -262, -0, 185, 0, -127, -0, 83, 0, -58, 0
            ]
            taps = 64
        # fmt: on
        current_rate = self._get_iio_attr("voltage0", "sampling_frequency", False)

        if self._get_iio_attr("out", "voltage_filter_fir_en", False):
            if current_rate <= (25000000 // 12):
                self._set_iio_attr("voltage0", "sampling_frequency", False, 3000000)
            self._set_iio_attr("out", "voltage_filter_fir_en", False, 0)

        # Assemble FIR filter config string
        fir_config_string = ""
        fir_config_string += "RX 3 GAIN -6 DEC " + str(dec) + "\n"
        fir_config_string += "TX 3 GAIN 0 INT " + str(dec) + "\n"
        for i in range(taps):
            fir_config_string += str(fir[i]) + "," + str(fir[i]) + "\n"
        fir_config_string += "\n"
        self._set_iio_dev_attr_str("filter_fir_config", fir_config_string)

        if rate <= (25000000 // 12):
            readbuf = self._get_iio_dev_attr_str("tx_path_rates")
            dacrate = int(readbuf.split(" ")[1].split(":")[1])
            txrate = int(readbuf.split(" ")[5].split(":")[1])
            max_rate = (dacrate // txrate) * 16
            if max_rate < taps:
                self._set_iio_attr("voltage0", "sampling_frequency", False, 3000000)
            self._set_iio_attr("out", "voltage_filter_fir_en", False, 1)
            self._set_iio_attr("voltage0", "sampling_frequency", False, rate)
        else:
            self._set_iio_attr("voltage0", "sampling_frequency", False, rate)
            self._set_iio_attr("out", "voltage_filter_fir_en", False, 1)

    @property
    def rx_lo(self):
        """rx_lo: Carrier frequency of RX path"""
        return self._get_iio_attr("altvoltage0", "frequency", True)

    @rx_lo.setter
    def rx_lo(self, value):
        self._set_iio_attr_int("altvoltage0", "frequency", True, value)

    @property
    def tx_lo(self):
        """tx_lo: Carrier frequency of TX path"""
        return self._get_iio_attr("altvoltage1", "frequency", True)

    @tx_lo.setter
    def tx_lo(self, value):
        self._set_iio_attr_int("altvoltage1", "frequency", True, value)


class ad9361(ad9364):
    """AD9361 Transceiver"""

    _rx_channel_names = ["voltage0", "voltage1", "voltage2", "voltage3"]
    _tx_channel_names = ["voltage0", "voltage1", "voltage2", "voltage3"]

    @property
    def gain_control_mode_chan1(self):
        """gain_control_mode_chan1: Mode of receive path AGC. Options are:
        slow_attack, fast_attack, manual"""
        return self._get_iio_attr_str("voltage1", "gain_control_mode", False)

    @gain_control_mode_chan1.setter
    def gain_control_mode_chan1(self, value):
        self._set_iio_attr("voltage1", "gain_control_mode", False, value)

    @property
    def rx_hardwaregain_chan1(self):
        """rx_hardwaregain_chan1: Gain applied to RX path. Only applicable when
        gain_control_mode is set to 'manual'"""
        return self._get_iio_attr("voltage1", "hardwaregain", False)

    @rx_hardwaregain_chan1.setter
    def rx_hardwaregain_chan1(self, value):
        if self.gain_control_mode_chan1 == "manual":
            self._set_iio_attr_float("voltage1", "hardwaregain", False, value)

    @property
    def tx_hardwaregain_chan1(self):
        """tx_hardwaregain_chan1: Attenuation applied to TX path"""
        return self._get_iio_attr("voltage1", "hardwaregain", True)

    @tx_hardwaregain_chan1.setter
    def tx_hardwaregain_chan1(self, value):
        self._set_iio_attr_float("voltage1", "hardwaregain", True, value)


class ad9363(ad9361):
    """AD9363 Transceiver"""

    pass


class Pluto(ad9364, _dec_int_fpga_filter):
    """PlutoSDR Evaluation Platform"""

    _device_name = "PlutoSDR"
    _uri_auto = "ip:pluto.local"

    def __repr__(self):
        retstr = f"""Pluto(uri="{self.uri}") object "{self._device_name}" with following key properties:

rx_lo:                   {self.rx_lo / 1000000:<12} MHz, Carrier frequency of RX path
rx_hardwaregain_chan0    {self.rx_hardwaregain_chan0:<12} dB, Gain applied to RX path. Only applicable when gain_control_mode is set to 'manual'
rx_rf_bandwidth:         {self.rx_rf_bandwidth / 1000000:<12} MHz, Bandwidth of front-end analog filter of RX path
gain_control_mode_chan0: {self.gain_control_mode_chan0:<12} Receive path AGC Options: slow_attack, fast_attack, manual

tx_lo:                   {self.tx_lo / 1000000:<12} MHz, Carrier frequency of TX path
tx_hardwaregain_chan0:   {self.tx_hardwaregain_chan0:<12} dB, Attenuation applied to TX path
tx_rf_bandwidth:         {self.tx_rf_bandwidth / 1000000:<12} MHz, Bandwidth of front-end analog filter of TX path
tx_cyclic_buffer:        {self.tx_cyclic_buffer:<12} Toggles cyclic buffer

filter:                  {str(self.filter):<12} FIR filter file
sample_rate:             {self.sample_rate / 1000000:<12} MSPS, Sample rate RX and TX paths
loopback:                {self.loopback:<12} 0=Disabled, 1=Digital, 2=RF

"""
        return retstr



# Copyright (C) 2023 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import adi


class cn0554:
    """ CN0554 Mixed Signal Raspberry Pi Hat """

    _in_range_available = ["+/-13.75", "+27.5", "+2.5"]
    _out_reference_available = [4.096, 2.5]
    _out_channels = 16
    _in_channels = 8
    _scale_res_div = 11

    def __init__(self, uri="ip:analog.local"):
        """adc_in_channels: ADC channel names derived from onboard AD7124."""
        self.adc_in_channels = []

        """dac_out_channels: DAC channel names derived from onboard LTC2688."""
        self.dac_out_channels = []

        self.adc = adi.ad7124(uri=uri)
        self.adc._rx_unbuffered_data = False
        self.adc._rx_output_type = "SI"
        self.rx_buffer_size = 1024
        self.adc.sample_rate = 19200

        for in_ch in self.adc.channel:
            in_name = "voltage"
            for char in in_ch.name:
                if char.isnumeric():
                    in_name += char

            """Default channels range is set to '+/-13.75'"""
            setattr(self, in_name + "_in_range", self._in_range_available[0])

            self.adc_in_channels.append(in_name)

        self.dac = adi.ltc2688(uri=uri)
        self.dac_out_channels = self.dac.channel_names

        """out_reference: DAC voltage reference in Volts. Valid values are 4.096 and 2.5."""
        self.out_reference = self._out_reference_available[0]

    @property
    def in_scale(self):
        """in_scale: unitless scale factor based on onboard resistor divider"""
        return self._scale_res_div

    @in_scale.setter
    def in_scale(self, value):
        """in_scale property setter"""
        if value <= 0:
            raise Exception("Scale Factor must be greater than 0")

        self._scale_res_div = value

    @property
    def rx_output_type(self):
        """Get value of CN0554's rx output type (raw or SI)"""
        return self.adc._rx_output_type

    @rx_output_type.setter
    def rx_output_type(self, value):
        """Set value of CN0554's rx output type (raw or SI)"""
        self.adc._rx_output_type = value

    @property
    def rx_enabled_channels(self):
        """Get list of enabled input adc channels"""
        return self.adc.rx_enabled_channels

    @rx_enabled_channels.setter
    def rx_enabled_channels(self, value):
        """Set list of enabled input adc channels"""
        self.adc.rx_enabled_channels = value

    @property
    def out_channels(self):
        """Get number of DAC output channels"""
        return self._out_channels

    @property
    def in_channels(self):
        """Get number of ADC input channels"""
        return self._in_channels

    @property
    def out_reference(self):
        """Get voltage reference used for DAC channels"""
        return self.dac.vref

    @out_reference.setter
    def out_reference(self, value):
        """Set voltage reference used for DAC channels"""
        if value not in self._out_reference_available:
            raise Exception("out_reference value must be 4.096 or 2.5")

        self.dac.vref = value

    @property
    def sample_rate(self):
        """Get CN0554's AD7124 sampling rate"""
        return self.adc.sample_rate

    @sample_rate.setter
    def sample_rate(self, value):
        """Set CN0554's AD7124 sampling rate"""
        self.adc.sample_rate = value

    @property
    def rx_buffer_size(self):
        """Get number of datapoints used during capture"""
        return self.adc.rx_buffer_size

    @rx_buffer_size.setter
    def rx_buffer_size(self, value):
        """Set number of datapoints used during capture"""
        self.adc.rx_buffer_size = value

    def rx(self):
        """Get data from enabled ADC channels"""
        return self.adc.rx()

    def convert_to_volts(self, in_voltage, channel):
        """Convert ADC data using scale factors based on header configuration"""
        ch_range = eval("self." + str(channel) + "_in_range")
        try:
            ch_index = self._in_range_available.index(ch_range)

        except ValueError:
            raise Exception(
                "Invalid _in_range detected. Valid values are '+/-13.75', '+27.5', and '+2.5'."
            )

        if ch_index < 2:
            scale_sig = []

            for sigs in in_voltage:
                scale_sig.append(sigs * self.in_scale)

            return scale_sig

        return in_voltage



# Copyright (C) 2020 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import datetime
import time
from typing import List

from adi.adrv9009_zu11eg import adrv9009_zu11eg
from adi.adrv9009_zu11eg_fmcomms8 import adrv9009_zu11eg_fmcomms8
from adi.jesd import jesd as jesd_api


class adrv9009_zu11eg_multi(object):
    """ADRV9009-ZU11EG Multi-SOM Manager

    parameters:
        primary_uri: type=string
            URI of primary ADRV9009-ZU11EG. Parent HMC7044 is connected
            to this SOM
        secondary_uris: type=list[string]
            URI(s) of secondary ADRV9009-ZU11EG(s).
        primary_jesd: type=adi.jesd
            JESD object associated with primary ADRV9009-ZU11EG
        secondary_jesds: type=list[adi.jesd]
            JESD object(s) associated with secondary ADRV9009-ZU11EG(s)
        fmcomms8: type=boolean
            Boolean flag to idenify is FMComms8(s) are attached to SOMs
    """

    __rx_buffer_size_multi = 2 ** 14
    secondaries: List[adrv9009_zu11eg] = []

    def __init__(
        self,
        primary_uri="",
        secondary_uris=[],
        primary_jesd=None,
        secondary_jesds=[None],
        fmcomms8=False,
    ):

        if not jesd_api:
            raise Exception(
                "JESD optional dependencies are required.\n"
                + "Please install them using pip install pyadi-iio[jesd] "
                + "or pip install paramiko"
            )

        if not isinstance(secondary_uris, list):
            Exception("secondary_uris must be a list")
        if not isinstance(secondary_jesds, list):
            Exception("secondary_jesds must be a list")

        self._dma_show_arming = False
        self._jesd_show_status = False
        self._jesd_fsm_show_status = False
        self._clk_chip_show_cap_bank_sel = False
        self._resync_tx = False
        self._rx_initialized = False
        self._request_sysref_carrier = False
        self.fmcomms8 = fmcomms8
        if fmcomms8:
            self.primary = adrv9009_zu11eg_fmcomms8(
                uri=primary_uri, jesd_monitor=True, jesd=primary_jesd
            )
        else:
            self.primary = adrv9009_zu11eg(
                uri=primary_uri, jesd_monitor=True, jesd=primary_jesd
            )
        self.secondaries = []
        self.samples_primary = []
        self.samples_secondary = []
        for i, uri in enumerate(secondary_uris):
            if fmcomms8:
                self.secondaries.append(
                    adrv9009_zu11eg_fmcomms8(
                        uri=uri, jesd_monitor=True, jesd=secondary_jesds[i]
                    )
                )
            else:
                self.secondaries.append(
                    adrv9009_zu11eg(uri=uri, jesd_monitor=True, jesd=secondary_jesds[i])
                )

        for dev in self.secondaries + [self.primary]:
            dev._rxadc.set_kernel_buffers_count(1)

    def reinitialize(self):
        """reinitialize: reinitialize all transceivers"""
        for dev in self.secondaries + [self.primary]:
            property_names = [p for p in dir(dev) if "reinitialize" in p]
            for p in property_names:
                eval("dev." + p + "()")

    @property
    def rx_buffer_size(self):
        """rx_buffer_size: Size of receive buffer in samples for each device"""
        return self.__rx_buffer_size_multi

    @rx_buffer_size.setter
    def rx_buffer_size(self, value):
        self.__rx_buffer_size_multi = value
        for dev in self.secondaries + [self.primary]:
            dev.rx_buffer_size = value

    def __read_jesd_status_all_devs(self, attr, islink=False):
        for dev in self.secondaries + [self.primary]:
            if islink:
                devs = dev._jesd.get_all_link_statuses()
                for dev in devs:
                    lanes = devs[dev]
                    print("JESD {}: ".format(dev), end="")
                    for lane in lanes:
                        if attr in lanes[lane]:
                            print(" {}".format(lanes[lane][attr]), end="")
                    print("")
            else:
                s = dev._jesd.get_all_statuses()
                for dev in s:
                    if attr in s[dev]:
                        print("JESD {}: {} ({})".format(attr, s[dev][attr], dev))

    def __read_jesd_status(self):
        self.__read_jesd_status_all_devs("Link status")
        self.__read_jesd_status_all_devs("SYSREF captured")
        self.__read_jesd_status_all_devs("SYSREF alignment error")

    def __read_jesd_link_status(self):
        self.__read_jesd_status_all_devs("Errors", True)
        self.__read_jesd_status_all_devs("Initial Lane Alignment Sequence", True)
        self.__read_jesd_status_all_devs("Initial Frame Synchronization", True)

    def __setup_framers(self):
        # adi,jesd204-framer-a-lmfc-offset 15
        for dev in self.secondaries + [self.primary]:
            v1 = (
                dev._get_iio_debug_attr_str(
                    "adi,jesd204-framer-a-lmfc-offset", dev._ctrl_b
                )
                == "15"
            )
            v2 = (
                dev._get_iio_debug_attr_str(
                    "adi,jesd204-framer-a-lmfc-offset", dev._ctrl
                )
                == "15"
            )
            if (not v1) or (not v2):
                dev._set_iio_debug_attr_str(
                    "adi,jesd204-framer-a-lmfc-offset", "15", dev._ctrl_b
                )
                dev._set_iio_debug_attr_str(
                    "adi,jesd204-framer-a-lmfc-offset", "15", dev._ctrl
                )
                dev._set_iio_debug_attr_str("initialize", "1", dev._ctrl_b)
                dev._set_iio_debug_attr_str("initialize", "1", dev._ctrl)
                print("Re-initializing JESD links")
                time.sleep(10)

    def _device_is_running(self, dev, index, verbose):
        err = dev.jesd204_fsm_error
        paused = dev.jesd204_fsm_paused
        state = dev.jesd204_fsm_state

        if verbose:
            print(
                "%s: DEVICE%d: Is <%s> in state <%s> with status <%d>"
                % (dev.uri, index, "Paused" if paused else "Running", state, err)
            )

        if err:
            print(
                "\nERROR %s: DEVICE%d: Is <%s> in state <%s> with status <%d>\n"
                % (dev.uri, index, "Paused" if paused else "Running", state, err)
            )
            return "error"

        state_last = state == "opt_post_running_stage"

        if (state_last == 0) and (paused == 0):
            return "running"

        if (state_last == 0) and (paused == 1):
            return "paused"

        if (state_last == 1) and (paused == 0):
            return "done"

        assert False

    def __jesd204_fsm_is_done(self):
        cnt = 0
        for index, dev in enumerate([self.primary] + self.secondaries):
            ret = self._device_is_running(dev, index, 0)
            if ret == "done":
                cnt += 1

        if cnt == len([self.primary] + self.secondaries):
            return "done"

    def _jesd204_fsm_sync(self):
        while True:
            if self.__jesd204_fsm_is_done() == "done":
                return "done"

            for index, dev in enumerate(self.secondaries + [self.primary]):
                ret = self._device_is_running(dev, index, self._jesd_fsm_show_status)
                if index == 0:
                    statep = dev.jesd204_fsm_state
                else:
                    state = dev.jesd204_fsm_state
                    assert state == statep

                if ret == "paused":
                    dev.jesd204_fsm_resume = "1"

                if ret == "error":
                    return ret

    def __unsync(self):
        for dev in [self.primary] + self.secondaries:
            dev._clock_chip.attrs["sleep_request"].value = "1"
            if self.fmcomms8:
                dev._clock_chip_fmc.attrs["sleep_request"].value = "1"
            dev._clock_chip_carrier.attrs["sleep_request"].value = "1"

        self.primary._clock_chip_ext.attrs["sleep_request"].value = "1"
        time.sleep(0.1)
        self.primary._clock_chip_ext.attrs["sleep_request"].value = "0"

        for dev in [self.primary] + self.secondaries:
            time.sleep(0.1)
            dev._clock_chip_carrier.attrs["sleep_request"].value = "0"
            time.sleep(0.1)
            if self.fmcomms8:
                dev._clock_chip_fmc.attrs["sleep_request"].value = "0"
            dev._clock_chip.attrs["sleep_request"].value = "0"

    def hmc7044_cap_sel(self):
        vals = []
        for dev in [self.primary] + self.secondaries:
            vals.append(dev._clock_chip.reg_read(0x8C))
            if self.fmcomms8:
                vals.append(dev._clock_chip_fmc.reg_read(0x8C))
            vals.append(dev._clock_chip_carrier.reg_read(0x8C))

        vals.append(self.primary._clock_chip_ext.reg_read(0x8C))
        return vals

    def hmc7044_set_cap_sel(self, vals):
        """hmc7044_set_cap_sel:

        parameters:
            vals: type=list
                Forces certain Capacitor bank selections.
                Typically the list returned form hmc7044_cap_sel
        """
        for dev in [self.primary] + self.secondaries:
            dev._clock_chip.reg_write(0xB2, vals.pop(0) << 2 | 1)
            if self.fmcomms8:
                dev._clock_chip_fmc.reg_write(0xB2, vals.pop(0) << 2 | 1)
            dev._clock_chip_carrier.reg_write(0xB2, vals.pop(0) << 2 | 1)

        self.primary._clock_chip_ext.reg_write(0xB2, vals.pop(0) << 2 | 1)

    def hmc7044_ext_output_delay(self, chan, digital, analog_ps):
        """hmc7044_ext_output_delay:

        parameters:
            digital: type=int
                Digital delay. Adjusts the phase of the divider signal
                by up to 17 half cycles of the VCO.
            analog_ps: type=int
                Analog delay. Adjusts the delay of the divider signal in
                increments of ~25 ps. Range is from 100ps to 700ps.
        """
        assert 0 <= chan <= 13
        if analog_ps - 100 >= 0:
            enable = 1
            val = (analog_ps - 100) / 25
        else:
            enable = 0
            val = 0

        offs = chan * 10
        self.primary._clock_chip_ext.reg_write(0xCF + offs, enable)
        self.primary._clock_chip_ext.reg_write(0xCB + offs, int(val) & 0x1F)
        self.primary._clock_chip_ext.reg_write(0xCC + offs, int(digital) & 0x1F)

    def hmc7044_car_output_delay(self, chan, digital, analog_ps):
        """hmc7044_car_output_delay:

        parameters:
            digital: type=int
                Digital delay. Adjusts the phase of the divider signal
                by up to 17 half cycles of the VCO.
            analog_ps: type=int
                Analog delay. Adjusts the delay of the divider signal in
                increments of ~25 ps. Range is from 100ps to 700ps.
        """
        assert 0 <= chan <= 13
        if analog_ps - 100 >= 0:
            enable = 1
            val = (analog_ps - 100) / 25
        else:
            enable = 0
            val = 0

        offs = chan * 10

        for dev in [self.primary] + self.secondaries:
            dev._clock_chip_carrier.reg_write(0xCF + offs, enable)
            dev._clock_chip_carrier.reg_write(0xCB + offs, int(val) & 0x1F)
            dev._clock_chip_carrier.reg_write(0xCC + offs, int(digital) & 0x1F)

    def __rx_dma_arm(self):
        for dev in self.secondaries + [self.primary]:
            if self._dma_show_arming:
                print("--DMA ARMING--", dev.uri)
            dev.rx_sync_start = "arm"
            if self._dma_show_arming:
                print("\n--DMA ARMED--", dev.uri)

    def __dds_sync_enable(self, enable):
        for dev in self.secondaries + [self.primary]:
            if self._dma_show_arming:
                print("--DAC SYNC ARMING--", dev.uri)
            dev.tx_sync_start = "arm"

    def sysref_request(self):
        """sysref_request: Sysref request for parent HMC7044"""
        if self._request_sysref_carrier:
            self.primary._clock_chip_carrier.attrs["sysref_request"].value = "1"
        else:
            self.primary._clock_chip_ext.attrs["sysref_request"].value = "1"

    def set_trx_lo_frequency(self, freq):
        """set_trx_lo_frequency:

        parameters:
            freq: type=int
                Frequency in hertz to be applied to all LOs
        """
        for dev in self.secondaries + [self.primary]:
            dev._set_iio_debug_attr_str("adi,trx-pll-lo-frequency_hz", freq, dev._ctrl)
            dev._set_iio_debug_attr_str(
                "adi,trx-pll-lo-frequency_hz", freq, dev._ctrl_b
            )
            if self.fmcomms8:
                dev._set_iio_debug_attr_str(
                    "adi,trx-pll-lo-frequency_hz", freq, dev._ctrl_c
                )
                dev._set_iio_debug_attr_str(
                    "adi,trx-pll-lo-frequency_hz", freq, dev._ctrl_d
                )

    def set_trx_framer_a_loopback(self, enable):
        """set_trx_framer_a_loopback: Set bist_framer_a_loopback
        """
        for dev in self.secondaries + [self.primary]:
            dev._set_iio_debug_attr_str("bist_framer_a_loopback", enable, dev._ctrl)
            dev._set_iio_debug_attr_str("bist_framer_a_loopback", enable, dev._ctrl_b)
            if self.fmcomms8:
                dev._set_iio_debug_attr_str(
                    "bist_framer_a_loopback", enable, dev._ctrl_c
                )
                dev._set_iio_debug_attr_str(
                    "bist_framer_a_loopback", enable, dev._ctrl_d
                )

    def __refill_samples(self, dev, is_primary):
        if is_primary:
            self.samples_primary = dev.rx()
        else:
            self.samples_secondary = dev.rx()

    def _pre_rx_setup(self):
        retries = 3
        for _ in range(retries):
            try:
                for dev in [self.primary] + self.secondaries:
                    dev.jesd204_fsm_ctrl = 0

                self.__unsync()

                for dev in [self.primary] + self.secondaries:
                    dev.jesd204_fsm_ctrl = 1

                self._jesd204_fsm_sync()

                if not self._resync_tx:
                    self.__dds_sync_enable(1)

                if self._clk_chip_show_cap_bank_sel:
                    print("HMC7044s CAP bank select: ", self.hmc7044_cap_sel())

                if self._jesd_show_status:
                    self.__read_jesd_status()
                    self.__read_jesd_link_status()

                for dev in [self.primary] + self.secondaries:
                    dev.rx_destroy_buffer()
                    dev._rx_init_channels()
                return
            except:  # noqa: E722
                print("Re-initializing due to lock-up")
                self.reinitialize()
        raise Exception("Unable to initialize (Board reboot required)")

    def rx(self):
        """Receive data from multiple hardware buffers for each channel index in
        rx_enabled_channels of each child object (primary,secondaries[indx]).

        returns: type=numpy.array or list of numpy.array
            An array or list of arrays when more than one receive channel
            is enabled containing samples from a channel or set of channels.
            Data will be complex when using a complex data device.
        """
        if not self._rx_initialized:
            self._pre_rx_setup()
            self._rx_initialized = True
        data = []
        self.__rx_dma_arm()
        # Recreate all buffers
        for dev in [self.primary] + self.secondaries:
            dev.rx_destroy_buffer()
            dev._rx_init_channels()

        if self._resync_tx:
            self.__dds_sync_enable(1)

        self.sysref_request()

        for dev in [self.primary] + self.secondaries:
            data += dev.rx()
        return data



# Copyright (C) 2021 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from adi.attribute import attribute
from adi.context_manager import context_manager


class yline(object):
    def __init__(self, dev, x, line):
        """Initializer for a Y line of switches."""
        self._line = line
        self._x = x
        self._dev = dev

    def __str__(self):
        """String representation of a Y line of switches."""
        return "[" + ", ".join([str(num) for num in self._line]) + "]"

    def __getitem__(self, y):
        """Read access a Y line using python array indexing."""
        return self._line[y]

    def __setitem__(self, y, value):
        """Write access a Y line using python array indexing."""
        self._dev._switch(self._x, y, value, self._dev._ldsw)
        self._line[y] = value


class adg2128(attribute, context_manager):
    """ADG2128 cross point switch."""

    def __init__(self, uri=""):
        """Initializer for the adg2128 cross point switch device."""
        self._device_name = "adg2128"
        context_manager.__init__(self, uri, self._device_name)

        self._i2c_devs = []
        self._xmax = 0
        self._xline = []
        self._ldsw = True

        self._ctrl = self._ctx.find_device(self._device_name)

        if not self._ctrl:
            raise Exception(self._device_name + " device not found")

    @property
    def immediate(self):
        """Specify whether the writing of a switch is immediate or not.

        When it's not immediate, the new switch configuration is only
        latched into the device (see LDSW in datasheet).
        """
        return self._ldsw

    @immediate.setter
    def immediate(self, value):
        self._ldsw = value

    def add(self, addr):
        """
        Add device by its i2c address.

        Multiple devices may be added provided they all have their Y terminals
        connected (common Y configuration).
        1x adg2128 is represented by a 12 by 8 matrix
        2x adg2128 is represented by a 24 by 8 matrix
        ...

        Arguments:
        addr - device address on i2c bus
        """
        self._i2c_devs.append(addr)
        self._xmax += 12
        for x in range(self._xmax - 12, self._xmax):
            y = yline(self, x, [False for y in range(8)])
            self._xline.append(y)

    def _read(self, addr):
        """Direct Register Access via debugfs."""
        self._set_iio_debug_attr_str("direct_reg_access", addr, self._ctrl)
        return self._get_iio_debug_attr_str("direct_reg_access", self._ctrl)

    def _write(self, addr, val):
        """Direct Register Access via debugfs."""
        self._set_iio_debug_attr_str(
            "direct_reg_access", "0x{:X} 0x{:X}".format(addr, val), self._ctrl
        )

    def _switch(self, x, y, closed, immediate):
        data = closed << 15
        ax_lookup = [0, 1, 2, 3, 4, 5, 8, 9, 10, 11, 12, 13]
        ax = ax_lookup[x % 12] << 11
        ay = y << 8
        ldsw = immediate
        addr = self._i2c_devs[x // 12]
        val = data | ax | ay | ldsw
        self._write(addr, val)

    def _read_x(self, x):
        addr = self._i2c_devs[int(x) // 12]
        x_lookup = [
            0b00110100,
            0b00111100,
            0b01110100,
            0b01111100,
            0b00110101,
            0b00111101,
            0b01110101,
            0b01111101,
            0b00110110,
            0b00111110,
            0b01110110,
            0b01111110,
        ]
        val = x_lookup[x % 12] << 8
        self._write(addr, val)
        return self._read(addr)

    def __str__(self):
        """String representation of the cross point switch."""
        return "\n".join([f"x{idx}: {x}" for idx, x in enumerate(self._xline)])

    def __getitem__(self, x):
        """Read access a X line using python array indexing."""
        return self._xline[x]

    def __setitem__(self, x, value):
        """Write access a X line using python array indexing."""
        for y, closed in enumerate(value):
            self._xline[x][y] = closed

    def open_all(self):
        """
        Open all switches.

        For each device, iterate all x-y combinations and
        open all the switches at once.
        """
        ldsw = self._ldsw
        for x in range(self._xmax):
            for y in range(8):
                self._ldsw = False
                if (y == 7) and (x % 12 == 11):
                    self._ldsw = True
                self[x][y] = False
        self._ldsw = ldsw



# Copyright (C) 2020 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from typing import Dict, List

from adi.ad9081 import ad9081
from adi.attribute import attribute
from adi.context_manager import context_manager
from adi.one_bit_adc_dac import one_bit_adc_dac
from adi.rx_tx import rx_tx
from adi.sync_start import sync_start


def _map_to_dict(paths, ch, dev_name):
    if "label" in ch.attrs and "buffer_only" in ch.attrs["label"].value:
        return paths, False
    fddc, cddc, adc = ch.attrs["label"].value.split("->")
    if dev_name not in paths.keys():
        paths[dev_name] = {}
    if adc not in paths[dev_name].keys():
        paths[dev_name][adc] = {}
    if cddc not in paths[dev_name][adc].keys():
        paths[dev_name][adc][cddc] = {}
    if fddc not in paths[dev_name][adc][cddc].keys():
        paths[dev_name][adc][cddc][fddc] = {"channels": [ch._id]}
    else:
        paths[dev_name][adc][cddc][fddc]["channels"].append(ch._id)
    return paths, True


def _sortconv(chans_names, noq=False, dds=False):
    tmpI = filter(lambda k: "_i" in k, chans_names)
    tmpQ = filter(lambda k: "_q" in k, chans_names)

    def ignoreadc(w):
        return int(w[len("voltage") : w.find("_")])

    def ignorealt(w):
        return int(w[len("altvoltage") :])

    chans_names_out = []
    if dds:
        filt = ignorealt
        tmpI = chans_names
        noq = True
    else:
        filt = ignoreadc

    tmpI = sorted(tmpI, key=filt)
    tmpQ = sorted(tmpQ, key=filt)
    for i in range(len(tmpI)):
        chans_names_out.append(tmpI[i])
        if not noq:
            chans_names_out.append(tmpQ[i])

    return chans_names_out


def _find_dev_with_buffers(ctx, output=False, contains=""):
    for dev in ctx.devices:
        for chan in dev.channels:
            if (
                chan.output == output
                and chan.scan_element
                and dev.name
                and contains in dev.name
            ):
                return dev


class ad9081_mc(ad9081):
    """AD9081 Mixed-Signal Front End (MxFE) Multi-Chip Interface

        This class is a generic interface for boards that utilize multiple AD9081
        devices.

    parameters:
        uri: type=string
            Optional parameter for the URI of IIO context with AD9081(s).
        phy_dev_name: type=string
            Optional parameter name of main control driver for multi-AD9081 board.
            If no argument is given the driver with the most channel attributes is
            assumed to be the main PHY driver

    """

    _complex_data = True
    _rx_channel_names: List[str] = []
    _tx_channel_names: List[str] = []
    _tx_control_channel_names: List[str] = []
    _rx_coarse_ddc_channel_names: List[str] = []
    _tx_coarse_duc_channel_names: List[str] = []
    _rx_fine_ddc_channel_names: List[str] = []
    _tx_fine_duc_channel_names: List[str] = []
    _dds_channel_names: List[str] = []
    _device_name = ""

    _rx_attr_only_channel_names: List[str] = []
    _tx_attr_only_channel_names: List[str] = []

    _path_map: Dict[str, Dict[str, Dict[str, List[str]]]] = {}

    def __init__(self, uri="", phy_dev_name=""):

        self._rx_channel_names: List[str] = []
        context_manager.__init__(self, uri, self._device_name)

        if not phy_dev_name:
            # Get ad9081 dev name with most channel attributes
            channel_attr_count = {
                dev.name: sum(len(chan.attrs) for chan in dev.channels)
                for dev in self._ctx.devices
                if dev.name and "ad9081" in dev.name
            }
            phy_dev_name = max(channel_attr_count, key=channel_attr_count.get)

        self._ctrl = self._ctx.find_device(phy_dev_name)
        if not self._ctrl:
            raise Exception("phy_dev_name not found with name: {}".format(phy_dev_name))

        # Find device with buffers
        self._txdac = _find_dev_with_buffers(self._ctx, True, "axi-ad9081")
        self._rxadc = _find_dev_with_buffers(self._ctx, False, "axi-ad9081")

        # Get DDC and DUC mappings
        # Labels span all devices so they must all be processed
        paths = {}
        self._default_ctrl_names = []
        for dev in self._ctx.devices:
            if dev.name and "ad9081" not in dev.name:
                continue
            for ch in dev.channels:
                not_buffer = False
                if "label" in ch.attrs:
                    paths, not_buffer = _map_to_dict(paths, ch, dev.name)
                if not_buffer and dev.name not in self._default_ctrl_names:
                    self._default_ctrl_names.append(dev.name)

        self._default_ctrl_names = sorted(self._default_ctrl_names)
        self._ctrls = [self._ctx.find_device(dn) for dn in self._default_ctrl_names]
        self._path_map = paths

        # Get data + DDS channels
        for ch in self._rxadc.channels:
            if ch.scan_element and not ch.output:
                self._rx_channel_names.append(ch._id)
        for ch in self._txdac.channels:
            if ch.scan_element:
                self._tx_channel_names.append(ch._id)
            else:
                self._dds_channel_names.append(ch._id)

        # Sort channel names
        self._rx_channel_names = _sortconv(self._rx_channel_names)
        self._tx_channel_names = _sortconv(self._tx_channel_names)
        self._dds_channel_names = _sortconv(self._dds_channel_names, dds=True)

        # Map unique attributes to channel properties
        self._map_unique(paths)

        # Bring up DMA and DDS interfaces
        rx_tx.__init__(self)
        sync_start.__init__(self)
        self.rx_buffer_size = 2 ** 16

    def _map_unique(self, paths):
        self._rx_fine_ddc_channel_names = {}
        self._rx_coarse_ddc_channel_names = {}
        self._tx_fine_duc_channel_names = {}
        self._tx_coarse_duc_channel_names = {}
        for chip in paths:
            for converter in paths[chip]:
                for cdc in paths[chip][converter]:
                    channels = []
                    for fdc in paths[chip][converter][cdc]:
                        channels += paths[chip][converter][cdc][fdc]["channels"]
                    channels = [name for name in channels if "_i" in name]

                    if "ADC" in converter:
                        if chip not in self._rx_coarse_ddc_channel_names.keys():
                            self._rx_coarse_ddc_channel_names[chip] = []
                        if chip not in self._rx_fine_ddc_channel_names.keys():
                            self._rx_fine_ddc_channel_names[chip] = []

                        self._rx_coarse_ddc_channel_names[chip].append(channels[0])
                        self._rx_fine_ddc_channel_names[chip] += channels
                    else:
                        if chip not in self._tx_coarse_duc_channel_names.keys():
                            self._tx_coarse_duc_channel_names[chip] = []
                        if chip not in self._tx_fine_duc_channel_names.keys():
                            self._tx_fine_duc_channel_names[chip] = []

                        self._tx_coarse_duc_channel_names[chip].append(channels[0])
                        self._tx_fine_duc_channel_names[chip] += channels

    def _map_inputs_to_dict(self, channel_names_dict, attr, output, values):
        if not isinstance(values, dict):
            # If passed an array it must be split across the devices
            # Check to make sure length is accurate
            d = self._get_iio_attr_vec(channel_names_dict, attr, output)
            t_len = sum(len(d[e]) for e in d)
            if len(values) != t_len:
                raise Exception("Input must be of length {}".format(t_len))
            h = {}
            c = 0
            for dev in self._default_ctrl_names:
                data_index = len(d[dev])
                h[dev] = values[c : c + data_index]
                c += data_index
            values = h
        return values

    def _map_inputs_to_dict_single(self, channel_names_dict, values):
        if not isinstance(values, dict):
            # If passed an array it must be split across the devices
            # Check to make sure length is accurate
            t_len = len(channel_names_dict)
            if len(values) != t_len:
                raise Exception("Input must be of length {}".format(t_len))
            h = {dev: values[i] for i, dev in enumerate(self._default_ctrl_names)}
            values = h
        return values

    # Vector function intercepts
    def _get_iio_attr_vec(self, channel_names_dict, attr, output):
        return {
            dev: ad9081._get_iio_attr_vec(
                self, channel_names_dict[dev], attr, output, self._ctx.find_device(dev),
            )
            for dev in channel_names_dict
        }

    def _set_iio_attr_int_vec(self, channel_names_dict, attr, output, values):
        values = self._map_inputs_to_dict(channel_names_dict, attr, output, values)
        for dev in channel_names_dict:
            ad9081._set_iio_attr_int_vec(
                self,
                channel_names_dict[dev],
                attr,
                output,
                values[dev],
                self._ctx.find_device(dev),
            )

    def _set_iio_attr_float_vec(self, channel_names_dict, attr, output, values):
        values = self._map_inputs_to_dict(channel_names_dict, attr, output, values)
        for dev in channel_names_dict:
            ad9081._set_iio_attr_float_vec(
                self,
                channel_names_dict[dev],
                attr,
                output,
                values[dev],
                self._ctx.find_device(dev),
            )

    def _set_iio_attr_str_vec(self, channel_names_dict, attr, output, values):
        values = self._map_inputs_to_dict(channel_names_dict, attr, output, values)
        for dev in channel_names_dict:
            ad9081._set_iio_attr_str_vec(
                self,
                channel_names_dict[dev],
                attr,
                output,
                values[dev],
                self._ctx.find_device(dev),
            )

    # Singleton function intercepts
    def _get_iio_attr_str_single(self, channel_name, attr, output):
        channel_names_dict = self._rx_coarse_ddc_channel_names
        return {
            dev: attribute._get_iio_attr_str(
                self, channel_name, attr, output, self._ctx.find_device(dev)
            )
            for dev in channel_names_dict
        }

    def _get_iio_attr_single(self, channel_name, attr, output):
        channel_names_dict = self._rx_coarse_ddc_channel_names
        return {
            dev: attribute._get_iio_attr(
                self, channel_name, attr, output, self._ctx.find_device(dev)
            )
            for dev in channel_names_dict
        }

    def _set_iio_attr_single(self, channel_name, attr, output, values):
        channel_names_dict = self._rx_coarse_ddc_channel_names
        values = self._map_inputs_to_dict_single(channel_names_dict, values)
        for dev in channel_names_dict:
            self._set_iio_attr(
                channel_name, attr, output, values[dev], self._ctx.find_device(dev)
            )

    def _get_iio_dev_attr_single(self, attr):
        channel_names_dict = self._rx_coarse_ddc_channel_names
        return {
            dev: attribute._get_iio_dev_attr(self, attr, self._ctx.find_device(dev))
            for dev in channel_names_dict
        }

    def _set_iio_dev_attr_single(self, attr, values):
        channel_names_dict = self._rx_coarse_ddc_channel_names
        values = self._map_inputs_to_dict_single(channel_names_dict, values)
        for dev in channel_names_dict:
            self._set_iio_dev_attr(attr, values[dev], self._ctx.find_device(dev))


class QuadMxFE(ad9081_mc):
    """Quad AD9081 Mixed-Signal Front End (MxFE) Development System

    parameters:
        uri: type=string
            Optional parameter for the URI of IIO context with QuadMxFE.
    """

    def __init__(self, uri="", calibration_board_attached=False):
        ad9081_mc.__init__(self, uri=uri, phy_dev_name="axi-ad9081-rx-3")
        one_bit_adc_dac.__init__(self, uri)

        self._clock_chip = self._ctx.find_device("hmc7043")

        self._rx_dsa = self._ctx.find_device("hmc425a")
        if not self._rx_dsa:
            self._rx_dsa = self._ctx.find_device("hmc540s")

        if calibration_board_attached:
            self._ad5592r = self._ctx.find_device("ad5592r")
            self._cb_gpio = self._ctx.find_device("one-bit-adc-dac")

    @property
    def rx_dsa_gain(self):
        """rx_dsa_gain: Receiver digital step attenuator gain"""
        return self._get_iio_attr("voltage0", "hardwaregain", True, self._rx_dsa)

    @rx_dsa_gain.setter
    def rx_dsa_gain(self, value):
        self._set_iio_attr("voltage0", "hardwaregain", True, value, self._rx_dsa)



# Copyright (C) 2021 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from adi.context_manager import context_manager
from adi.rx_tx import rx


class ad6676(rx, context_manager):
    """ AD6676 Wideband IF Receiver Subsystem """

    _complex_data = True
    _rx_channel_names = ["voltage0", "voltage1"]
    _device_name = ""

    def __init__(self, uri=""):

        context_manager.__init__(self, uri, self._device_name)

        self._rxadc = self._ctx.find_device("axi-ad6676-hpc")
        self._ctrl = self._rxadc

        rx.__init__(self)

    @property
    def adc_frequency(self):
        """adc_frequency: The clock frequency of the ADC. Maximizing the clock frequency is helpful
        when the IF or bandwidth are high. A lower clock frequency reduces power consumption and is
        appropriate for low IFs and narrow bandwidths.

        Range using external synthesizer [2.0,3.2] GHz in Hz
        Range using internal synthesizer [2.925,3.2] GHz in Hz"""
        return self._get_iio_attr_str("voltage0", "adc_frequency", False)

    @adc_frequency.setter
    def adc_frequency(self, value):
        self._set_iio_attr("voltage0", "adc_frequency", False, value)

    @property
    def bandwidth(self):
        """bandwidth: The bandwidth of the ADC. Since the AD6676 uses delta-sigma technology,
        the available bandwidth is a relatively small fraction of the ADC clock rate and
        the AD6676 achieves the lowest noise and distortion when the bandwidth is small.

        The allowed range is between [0.005,0.05]*FADC in Hz"""
        return self._get_iio_attr_str("voltage0", "bandwidth", False)

    @bandwidth.setter
    def bandwidth(self, value):
        self._set_iio_attr("voltage0", "bandwidth", False, value)

    @property
    def bw_margin_high(self):
        """bw_margin_high: High/upper bandwidth margins for the noise-shaping profile of the ADC.
        Typical values are 5 MHz, but the user may want to increase these margins in wideband operation
        in order to optimize the noise profile of the ADC.
        Typical range [0,30] MHz in MHz"""
        return self._get_iio_attr_str("voltage0", "bw_margin_high", False)

    @bw_margin_high.setter
    def bw_margin_high(self, value):
        self._set_iio_attr("voltage0", "bw_margin_high", False, value)

    @property
    def bw_margin_if(self):
        """bw_margin_if: Displacement of the resonance frequency (F1Shift) of the first resonator within
        the ADC from band-center. Typically 0 is appropriate, but in the widest bandwidth modes
        positive shifts can be used to reduce the noise density near the upper edge of the passband.
        Typical range [-30,30] MHz in MHz"""
        return self._get_iio_attr_str("voltage0", "bw_margin_if", False)

    @bw_margin_if.setter
    def bw_margin_if(self, value):
        self._set_iio_attr("voltage0", "bw_margin_if", False, value)

    @property
    def bw_margin_low(self):
        """bw_margin_low: Lower bandwidth margins for the noise-shaping profile of the ADC.
        Typical values are 5 MHz, but the user may want to increase these margins in wideband operation
        in order to optimize the noise profile of the ADC.
        Typical range [0,30] MHz in MHz"""
        return self._get_iio_attr_str("voltage0", "bw_margin_low", False)

    @bw_margin_low.setter
    def bw_margin_low(self, value):
        self._set_iio_attr("voltage0", "bw_margin_low", False, value)

    @property
    def hardwaregain(self):
        """hardwaregain: The AD6676 contains a 50-ohm input attenuator programmable in 1-dB steps.
        Use this device attribute to set the attenuator's attenuation.
        The dynamic range of the system increases somewhat with moderate attenuation settings of 6-12 dB,
        at the expense of an increased noise figure.
        The range is from 0 to -27.00 dB in 1dB steps.
        The nomenclature used here is gain instead of attenuation, so all values are expressed negative."""
        return self._get_iio_attr_str("voltage0", "hardwaregain", False)

    @hardwaregain.setter
    def hardwaregain(self, value):
        self._set_iio_attr("voltage0", "hardwaregain", False, value)

    @property
    def scale(self):
        """scale: One of the convenient features of the AD6676 is that the full-scale of its ADC is
        relatively small and adjustable over a 12-dB range [1.00 .. 0.25]
        The dynamic range of the ADC is highest at the maximum full-scale setting but the noise
        figure of the system is lowest at the minimum full-scale setting.

        Writing a value of 0.5 to this device attribute lowers the PIN_0dBFS by 6 dB.
        Likewise writing a value of 0.25 to this device attribute lowers the PIN_0dBFS by 12 dB."""
        return self._get_iio_attr_str("voltage0", "scale", False)

    @scale.setter
    def scale(self, value):
        self._set_iio_attr("voltage0", "scale", False, value)

    @property
    def intermediate_frequency(self):
        """intermediate_frequency: The IF (intermediate frequency) to which the ADC is tuned.
        The AD6676 supports IFs from 70 to 450 MHz provided the external inductors are chosen appropriately.
        Since the AD6676-EBZ by default includes a pair of 19-nH inductors soldered to the evaluation board,
        the IF range allowed is less than the full range supported by the AD6676."""
        return self._get_iio_attr_str("voltage0", "intermediate_frequency", False)

    @intermediate_frequency.setter
    def intermediate_frequency(self, value):
        self._set_iio_attr("voltage0", "intermediate_frequency", False, value)

    @property
    def sampling_frequency(self):
        """sampling_frequency: The complex (I/Q) data rate in SPS.
        The AD6676 supports decimation factors (DFs) of 12, 16, 24 and 32.
        The complex (I/Q) data rate at the JESD204B outputs is FADC / DF."""
        return self._get_iio_attr_str("voltage0", "sampling_frequency", False)

    @sampling_frequency.setter
    def sampling_frequency(self, value):
        self._set_iio_attr("voltage0", "sampling_frequency", False, value)

    @property
    def shuffler_control(self):
        """shuffler_control: The AD6676 includes dynamic reordering of the comparators within the ADC in order to
        break up the spurious tones and distortion products associated with a fixed ordering.
        The Shuffle Control device attribute allows the user to experiment with different shuffling rates.
        The 'fadc' option (Shuffle every 1) reorders the comparators on every clock cycle with 50% probability.
        This shuffle scheme is able to randomize deterministic spurs but tends to increase the
        noise density and creates FADC/32 "shuffle humps" in the output spectrum.
        Similarly, the 'fadc/2' 'fadc/3' 'fadc/4' (Shuffle every 2,3,4) options reorder the comparators every n clock cycles
        with 50% probability. Using a high value of n decreases the noise degradation at the
        expense of less effective randomization and FADC/(32*n) shuffle humps that are closer
        to the main carrier. Fast shuffling can be disabled by selecting the **disable** option.
        Available values: disable fadc fadc/2 fadc/3 fadc/4"""
        return self._get_iio_attr_str("voltage0", "shuffler_control", False)

    @shuffler_control.setter
    def shuffler_control(self, value):
        self._set_iio_attr("voltage0", "shuffler_control", False, value)

    @property
    def shuffler_thresh(self):
        """shuffler_thresh: In order to obtain the spur-reduction benefits of shuffling at large signal levels while
        retaining the low noise of not shuffling when the signal is small, the AD6676 supports
        dynamic shuffle control via the Shuffle Threshold attribute.
        Shuffling is disabled if the raw ADC output is below the specified threshold for ~5000 clock cycles.
        A threshold of zero implies that shuffling is always enabled.
        The supported range is from 0..8"""
        return self._get_iio_attr_str("voltage0", "shuffler_thresh", False)

    @shuffler_thresh.setter
    def shuffler_thresh(self, value):
        self._set_iio_attr("voltage0", "shuffler_thresh", False, value)

    @property
    def test_mode(self):
        """test_mode: Select Test Mode. Options are:
        off midscale_short pos_fullscale neg_fullscale checkerboard pn_long pn_short one_zero_toggle user ramp"""
        return self._get_iio_attr("voltage0", "test_mode", False)

    @test_mode.setter
    def test_mode(self, value):
        self._set_iio_attr("voltage0", "test_mode", False, value, self._rxadc)



# Copyright (C) 2019 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from adi.ad469x import ad469x
from adi.ad717x import ad717x
from adi.ad719x import ad719x
from adi.ad777x import ad777x
from adi.ad936x import Pluto, ad9361, ad9363, ad9364
from adi.ad4020 import ad4020
from adi.ad4110 import ad4110
from adi.ad4130 import ad4130
from adi.ad4630 import ad4630
from adi.ad5592r import ad5592r
from adi.ad5686 import ad5686
from adi.ad5940 import ad5940
from adi.ad6676 import ad6676
from adi.ad7124 import ad7124
from adi.ad7606 import ad7606
from adi.ad7689 import ad7689
from adi.ad7746 import ad7746
from adi.ad7768 import ad7768
from adi.ad7799 import ad7799
from adi.ad9081 import ad9081
from adi.ad9081_mc import QuadMxFE, ad9081_mc
from adi.ad9083 import ad9083
from adi.ad9094 import ad9094
from adi.ad9136 import ad9136
from adi.ad9144 import ad9144
from adi.ad9152 import ad9152
from adi.ad9162 import ad9162
from adi.ad9166 import ad9166
from adi.ad9172 import ad9172
from adi.ad9250 import ad9250
from adi.ad9265 import ad9265
from adi.ad9371 import ad9371
from adi.ad9434 import ad9434
from adi.ad9467 import ad9467
from adi.ad9625 import ad9625
from adi.ad9680 import ad9680
from adi.ada4961 import ada4961
from adi.adaq8092 import adaq8092
from adi.adar1000 import adar1000, adar1000_array
from adi.adf4159 import adf4159
from adi.adf4355 import adf4355
from adi.adf4371 import adf4371
from adi.adf5610 import adf5610
from adi.adg2128 import adg2128
from adi.adis16460 import adis16460
from adi.adis16495 import adis16495
from adi.adis16507 import adis16507
from adi.adl5240 import adl5240
from adi.adl5960 import adl5960
from adi.admv8818 import admv8818
from adi.adpd188 import adpd188
from adi.adpd410x import adpd410x
from adi.adpd1080 import adpd1080
from adi.adrf5720 import adrf5720
from adi.adrv9002 import adrv9002
from adi.adrv9009 import adrv9009
from adi.adrv9009_zu11eg import adrv9009_zu11eg
from adi.adrv9009_zu11eg_fmcomms8 import adrv9009_zu11eg_fmcomms8
from adi.adrv9009_zu11eg_multi import adrv9009_zu11eg_multi
from adi.adt7420 import adt7420
from adi.adxl313 import adxl313
from adi.adxl345 import adxl345
from adi.adxl355 import adxl355
from adi.adxrs290 import adxrs290
from adi.cn0511 import cn0511
from adi.cn0532 import cn0532
from adi.cn0554 import cn0554
from adi.daq2 import DAQ2
from adi.daq3 import DAQ3
from adi.fmc_vna import fmcvna
from adi.fmcadc3 import fmcadc3
from adi.fmcjesdadc1 import fmcjesdadc1
from adi.fmclidar1 import fmclidar1
from adi.fmcomms5 import FMComms5
from adi.fmcomms11 import FMComms11
from adi.gen_mux import genmux
from adi.lm75 import lm75
from adi.ltc2314_14 import ltc2314_14
from adi.ltc2387 import ltc2387
from adi.ltc2499 import ltc2499
from adi.ltc2688 import ltc2688
from adi.ltc2983 import ltc2983
from adi.max9611 import max9611
from adi.max11205 import max11205
from adi.max31855 import max31855
from adi.max31865 import max31865
from adi.one_bit_adc_dac import one_bit_adc_dac
from adi.QuadMxFE_multi import QuadMxFE_multi
from adi.tdd import tdd

try:
    from adi.jesd import jesd
except ImportError:
    pass

__version__ = "0.0.15"
name = "Analog Devices Hardware Interfaces"



# Copyright (C) 2020 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from adi.rx_tx import rx, tx


class _property_remap:
    def __init__(self, attr, dev):
        self._dev = dev
        self._attr = attr

    def __get__(self, instance, owner):
        return getattr(self._dev, self._attr)

    def __set__(self, instance, value):
        setattr(self._dev, self._attr, value)


def remap(object_source, old, new, classname):
    """ Remap methods to new names to different object """

    method_list = [
        func
        for func in dir(object_source)
        if not func.startswith("__") and not func.startswith("_")
    ]
    for method in method_list:
        if old not in method:
            continue
        new_method = method.replace(old, new)
        setattr(
            classname, new_method, _property_remap(method, object_source),
        )


class obs(rx):
    """ Buffer handling for observation devices """

    def __init__(self, ctx, obs_dev, channel_names, complex_data=True):
        self._ctx = ctx
        self._rxadc = obs_dev
        self._rx_channel_names = channel_names
        self._complex_data = complex_data
        rx.__init__(self)

    def __del__(self):
        rx.__del__(self)


class tx_two(tx):
    """ Buffer Handling for secondary transmit channels """

    def __init__(self, ctx, obs_dev, channel_names, complex_data=True):
        self._ctx = ctx
        self._txdac = obs_dev
        self._tx_channel_names = channel_names
        self._complex_data = complex_data
        tx.__init__(self)

    def __del__(self):
        tx.__del__(self)
